from spacy.matcher import Matcher
from spacy.tokens import Doc


class CompoundCases:
    """
    Detects compound references to cases. Two styles of referencing are
    accommodated: common_pattern (e.g. Smith v Jones [2000] 1 WLR 123) and
    the slightly old fashioned possessive_pattern (e.g in Hoadley's case 
    [2018] 1 WLR 123).

    This class sets the `._.compound_cases` attribute on the spaCy doc.

    Note! This class merges all of the entities in the doc and replaces
    the spans.

    Usage:

    compound_pipe = CompoundCases(nlp)
    nlp.add_pipe(compound_pipe)

    for compound in doc._.compound_cases:
        ...
    """

    def __init__(self, nlp) -> None:
        Doc.set_extension("compound_cases", default=[], force=True)
        self.matcher = Matcher(nlp.vocab)
        common_pattern = [{"ent_type": "CASENAME"}, {"ent_type": "CITATION", "OP": "+"}]
        possessive_pattern = [
            {"ent_type": "CASENAME"},
            {"lower": "case"},
            {"ent_type": "CITATION"},
        ]
        self.matcher.add("compound_case", None, common_pattern, possessive_pattern)
        self.global_matcher = Matcher(nlp.vocab)
        merge_ents = nlp.create_pipe("merge_entities")
        nlp.add_pipe(merge_ents)

    def __call__(self, doc: Doc) -> Doc:

        compound_cases = self.matcher(doc)
        for match_id, start, end in compound_cases:
            span = doc[start:end]  # The matched span
            doc._.compound_cases.append(span.text)

        return doc



from spacy.matcher import Matcher

class SentenceSegmenter(object):
    """Adds custom sentence boundaries to spaCy Doc. 
    @param vocab: the language vocabulary from the spaCy nlp object
    @param boundary_protection_rules: an optional list of labelled matching patterns that will explicitly prevent sentence boundaries
    """
    name = "sentence_segmenter"

    def __init__(self, vocab, boundary_protection_rules=[]):
        self.matcher = Matcher(vocab)
        for rule in boundary_protection_rules:
            self.matcher.add(rule['label'], None, rule['pattern'])

    def __call__(self, doc):
        matches = self.matcher(doc)
        for i, token in enumerate(doc[:-2]):
            if token_in_match_range(token, matches):
                token.is_sent_start = False
            if token.text == "Sch" and doc[i+1].is_punct:
                doc[i+2].is_sent_start = False
            if str(doc[i]) == "]" and doc[i+1].is_digit:
                doc[i+2].is_sent_start = False
            if str(doc[i+1]) == ")":
                doc[i+2].is_sent_start = False
            if str(doc[i+1]) == "(":
                doc[i+2].is_sent_start = False
            if str(doc[i+1]) == "]":
                doc[i+2].is_sent_start = False
            if str(doc[i+1]) == "[":
                doc[i+2].is_sent_start = False
            if doc[i+1].is_digit:
                doc[i+2].is_sent_start = False
            if str(doc[i+1]) == ";":
                doc[i+2].is_sent_start = False
            if str(doc[i+1]) == ":":
                doc[i+2].is_sent_start = True
            if str(doc[i+1]) == "\n\n":
                doc[i+2].is_sent_start = True
            if str(doc[i+1]) == "\n":
                doc[i+2].is_sent_start = True
            if str(doc[i+1]) == "-":    
                doc[i+2].is_sent_start = False
            if str(doc[i+2]) == "[":
                doc[i+2].is_sent_start = False
            if doc[i].is_digit and str(doc[i+1]) == "]":
                doc[i+1].is_sent_start = False
            if doc[i].is_digit:
                doc[i+1].is_sent_start = False
            if token.text == "?" and doc[i+1].is_quote:
                doc[i+1].is_sent_start = False
            if token.text == "ex":
                doc[i].is_sent_start = False
            if token.text == "ex" and str(doc[i+1]) == "parte":
                doc[i+2].is_sent_start = False
            if token.text == "\n" and str(doc[i+1]) == "\n":
                doc[i+2].is_sent_start = True
            if token.text == "cf" and str(doc[i+1]) == ".":
                doc[i+2].is_sent_start = False
            if token.text == "unreported" and str(doc[i+1]) == ":":
                doc[i+2].is_sent_start = False
            if token.text == "Hon" and str(doc[i+1]) == ".":
                doc[i+2].is_sent_start = False
            if token.text == "." and doc[i+1].is_quote:
                doc[i+1].is_sent_start = False
            if token.text == "?" and str(doc[i+1]) ==   "!":
                doc[i+1].is_sent_start = False
            if token.text == "!" and str(doc[i+1]) == "?":
                doc[i+1].is_sent_start = False
            if str(doc[i+2]) == "(":
                doc[i+2].is_sent_start = False
            if token.text == "reg" and doc[i+1].is_punct:
                doc[i+2].is_sent_start = False
            if token.text == "’s":
                doc[i].is_sent_start = False
            if doc[i].is_quote:
                doc[i+1].is_sent_start = False
            if token.text == ")" and doc[i+1].is_punct == False:
                doc[i+2].is_sent_start = False
            if token.text == "Eu" and doc[i+1].is_punct:
                doc[i+2].is_sent_start = False
        return doc
    
def token_in_match_range(token, matches):
    for _, start, end in matches:
        if token.i >= start and token.i <= end:
            return True
    return False

